"""Utility functions for hyfi"""
import ast
import datetime
import itertools
import os
import re
import time
from contextlib import contextmanager
from functools import partial
from timeit import default_timer
from typing import List, Type

import chardet


def unescape_dict(d):
    """Unescape a dictionary"""
    return ast.literal_eval(repr(d).encode("utf-8").decode("unicode-escape"))


def _elapser_timer(start):
    """A function that returns the elapsed time"""
    return default_timer() - start


def _elapser(start, end):
    """A function that returns the elapsed time"""
    return end - start


@contextmanager
def elapsed_timer(format_time=False):
    """A context manager that yields a function that returns the elapsed time"""
    start = default_timer()
    # elapser = lambda: default_timer() - start
    elapser = partial(_elapser_timer, start)
    yield lambda: str(
        datetime.timedelta(seconds=elapser())
    ) if format_time else elapser()
    end = default_timer()
    # elapser = lambda: end - start
    elapser = partial(_elapser, start, end)


def lower_case_with_underscores(string):
    """Converts 'CamelCased' to 'camel_cased'."""
    return re.sub(r"\s+", "_", string.lower()).replace("-", "_")


def ordinal(num):
    """Return the ordinal of a number as a string."""
    return "%d%s" % (
        num,
        "tsnrhtdd"[(num // 10 % 10 != 1) * (num % 10 < 4) * num % 10 :: 4],
    )


def get_offset_ranges(count, num_workers):
    """Get offset ranges for parallel processing"""
    assert count > num_workers
    step_sz = int(count / num_workers)
    offset_ranges = [0]
    pv_cnt = 1
    for i in range(num_workers):
        pv_cnt = count if i == num_workers - 1 else pv_cnt + step_sz
        offset_ranges.append(pv_cnt)
    return offset_ranges


@contextmanager
def change_directory(directory):
    """Change directory and change back to original directory"""
    original = os.path.abspath(os.getcwd())

    fancy_print(f" Change directory to {directory}")
    os.chdir(directory)
    try:
        yield

    except Exception as e:
        fancy_print(f" Exception: {e}")
        raise e

    finally:
        fancy_print(f" Change directory back to {original}")
        os.chdir(original)


def fancy_print(*args, color=None, bold=False, **kwargs):
    """Print with color and bold"""
    if bold:
        print("\033[1m", end="")

    if color:
        print(f"\033[{color}m", end="")

    print(*args, **kwargs)

    print("\033[0m", end="")  # reset


# https://stackoverflow.com/questions/12523586/python-format-size-application-converting-b-to-kb-mb-gb-tb/37423778
def humanbytes(B, units=None):
    "Return the given bytes as a human friendly KB, MB, GB, or TB string"
    B = float(B)
    KB = float(1024)
    MB = float(KB**2)  # 1,048,576
    GB = float(KB**3)  # 1,073,741,824
    TB = float(KB**4)  # 1,099,511,627,776

    if (B < KB and units is None) or units == "B":
        return "{0} {1}".format(B, "Bytes" if 0 == B > 1 else "Byte")
    elif (KB <= B < MB and units is None) or units == "KiB":
        return "{0:.2f} KiB".format(B / KB)
    elif (MB <= B < GB and units is None) or units == "MiB":
        return "{0:.2f} MiB".format(B / MB)
    elif (GB <= B < TB and units is None) or units == "GiB":
        return "{0:.2f} GiB".format(B / GB)
    elif (TB <= B and units is None) or units == "TiB":
        return "{0:.2f} TiB".format(B / TB)


def parse_size(sizestr):
    """
    Parse a size string into a number of bytes. For example, "16K" will
    return 16384.  If no suffix is provided, bytes are assumed.  This
    function is case-insensitive.

    :param sizestr: A string representing a size, such as "16K", "2M", "1G".
    :return: The number of bytes that the string represents.
    """
    unit = sizestr[-1]
    size = float(sizestr[:-1])

    if unit.upper() == "B":
        return size
    if unit.upper() == "K":
        return size * 1024
    if unit.upper() == "M":
        return size * 1024 * 1024
    if unit.upper() == "G":
        return size * 1024 * 1024 * 1024
    if unit.upper() == "T":
        return size * 1024 * 1024 * 1024 * 1024


def check_min_len(s, len_func, min_len):
    """Check if the length of a string is greater than or equal to a minimum length"""
    return len_func(s) >= min_len


def check_max_len(s, len_func, max_len):
    """Check if the length of a string is less than or equal to a maximum length"""
    return len_func(s) <= max_len


def utf8len(s):
    """Return the length of a string in bytes"""
    return len(str(s).encode("utf-8"))


def len_wospc(x):
    """Return the length of a string in bytes without spaces"""
    return utf8len(re.sub(r"\s", "", str(x)))


def len_bytes(x):
    """Return the length of a string in bytes"""
    return utf8len(x)


def len_words(x):
    """Return the number of words in a string"""
    return len(x.split()) if isinstance(x, str) else 0


def len_sents(x, sep):
    """Return the number of sentences in a string"""
    sep = str(sep).encode("utf-8").decode("unicode-escape")
    return len(re.sub(r"(\r?\n|\r){1,}", sep, x).split(sep))


def len_segments(x, sep):
    """Return the number of segments in a string"""
    sep = str(sep).encode("utf-8").decode("unicode-escape")
    return len(re.sub(r"(\r?\n|\r){2,}", sep, x).split(sep))


def any_to_utf8(b):
    """Convert any string to utf-8"""
    try:
        return b.decode("utf-8")
    except UnicodeDecodeError:
        # try to figure out encoding if not utf-8

        guess = chardet.detect(b)["encoding"]

        if not guess or guess == "UTF-8":
            return

        try:
            return b.decode(guess)
        except (UnicodeDecodeError, LookupError):
            # still cant figure out encoding, give up
            return


def get_modified_time(path):
    """Return the modification time of a file"""
    if not os.path.exists(path):
        return None
    modTimesinceEpoc = os.path.getmtime(path)
    return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(modTimesinceEpoc))


def today(_format="%Y-%m-%d"):
    """Return today's date"""
    from datetime import datetime

    if _format is None:
        return datetime.today().date()
    else:
        return datetime.today().strftime(_format)


def now(_format="%Y-%m-%d %H:%M:%S"):
    """Return current date and time"""
    from datetime import datetime

    return datetime.now() if _format is None else datetime.now().strftime(_format)


def strptime(
    _date_str: str,
    _format: str = "%Y-%m-%d",
):
    """Return a datetime object from a string"""
    from datetime import datetime

    return datetime.strptime(_date_str, _format)


def to_dateparm(_date, _format="%Y-%m-%d"):
    """Return a date parameter string"""
    from datetime import datetime

    _dtstr = datetime.strftime(_date, _format)
    _dtstr = "${to_datetime:" + _dtstr + "," + _format + "}"
    return _dtstr


def to_datetime(data, _format=None, _columns=None, **kwargs):
    """Convert a string, int, or datetime object to a datetime object"""
    from datetime import datetime

    import pandas as pd

    if isinstance(data, datetime):
        return data
    elif isinstance(data, str):
        if _format is None:
            _format = "%Y-%m-%d"
        return datetime.strptime(data, _format)
    elif isinstance(data, int):
        return datetime.fromtimestamp(data)
    elif isinstance(data, pd.DataFrame):
        if _columns:
            if isinstance(_columns, str):
                _columns = [_columns]
            for _col in _columns:
                data[_col] = pd.to_datetime(data[_col], format=_format, **kwargs)
        return data
    else:
        return data


def to_numeric(data, _columns=None, errors="coerce", downcast=None, **kwargs):
    """Convert a string, int, or float object to a float object"""
    import pandas as pd

    if isinstance(data, str):
        return float(data)
    elif isinstance(data, (int, float)) or not isinstance(data, pd.DataFrame):
        return data
    else:
        if _columns:
            if isinstance(_columns, str):
                _columns = [_columns]
            for _col in _columns:
                data[_col] = pd.to_numeric(data[_col], errors=errors, downcast=downcast)
        return data


def human_readable_type_name(t: Type) -> str:
    """
    Generates a useful-for-humans label for a type.
    For builtin types, it's just the class name (eg "str" or "int").
    For other types, it includes the module (eg "pathlib.Path").
    """
    module = t.__module__
    if module == "builtins":
        return t.__qualname__
    elif module.split(".")[0] == "hyfi":
        module = "hyfi"

    try:
        return f"{module}.{t.__qualname__}"
    except AttributeError:
        return str(t)


def readable_types_list(type_list: List[Type]) -> str:
    """Generates a useful-for-humans label for a list of types."""
    return ", ".join(human_readable_type_name(t) for t in type_list)


def dict_product(dicts):
    """
    >>> list(dict_product(dict(number=[1,2], character='ab')))
    [{'character': 'a', 'number': 1},
     {'character': 'a', 'number': 2},
     {'character': 'b', 'number': 1},
     {'character': 'b', 'number': 2}]
    """
    return (dict(zip(dicts, x)) for x in itertools.product(*dicts.values()))


def dict_to_dataframe(data, orient="columns", dtype=None, columns=None):
    """Convert a dictionary to a pandas dataframe"""
    import pandas as pd

    return pd.DataFrame.from_dict(data, orient=orient, dtype=dtype, columns=columns)


def records_to_dataframe(
    data, index=None, exclude=None, columns=None, coerce_float=False, nrows=None
):
    """Convert a list of records to a pandas dataframe"""
    import pandas as pd

    return pd.DataFrame.from_records(
        data,
        index=index,
        exclude=exclude,
        columns=columns,
        coerce_float=coerce_float,
        nrows=nrows,
    )



import io
from pathlib import Path
from typing import Optional

from beaker import Beaker, ChecksumFailedError, DatasetNotFound, DatasetReadError

from .scheme_client import SchemeClient


class BeakerClient(SchemeClient):
    scheme = ("beaker",)
    recoverable_errors = SchemeClient.recoverable_errors + (
        DatasetReadError,
        ChecksumFailedError,
    )

    def __init__(self, resource: str) -> None:
        super().__init__(resource)
        self.beaker = Beaker.from_env()
        # Beaker resources should be in the form "{user}/{dataset_name}/{path}/{to}/{file}"
        path = Path(resource.split("://")[1])
        if len(path.parts) < 2:
            raise ValueError(
                f"Invalid beaker resource URL '{resource}'. "
                "Resources should be in the form 'beaker://{user_name}/{dataset_name}/{path_to_file}' "
                "or beaker://{dataset_id}/{path_to_file}."
            )

        try:
            user, dataset_name, *filepath_parts = path.parts
            self.dataset = self.beaker.dataset.get(f"{user}/{dataset_name}")
        except DatasetNotFound:
            dataset_id, *filepath_parts = path.parts
            self.dataset = self.beaker.dataset.get(dataset_id)

        self.filepath = "/".join(filepath_parts)
        self.file_info = self.beaker.dataset.file_info(self.dataset, self.filepath)

    def get_etag(self) -> Optional[str]:
        return None if self.file_info.digest is None else str(self.file_info.digest)

    def get_size(self) -> Optional[int]:
        return self.file_info.size

    def get_resource(self, temp_file: io.BufferedWriter) -> None:
        for chunk in self.beaker.dataset.stream_file(
            self.dataset, self.filepath, quiet=True
        ):
            if chunk:
                temp_file.write(chunk)



import os
from pathlib import Path

import gdown

from hyfi.cached_path._cached_path import _cached_path as _cpath
from hyfi.utils.logging import getLogger

logger = getLogger(__name__)


def cached_path(
    url_or_filename,
    extract_archive: bool = False,
    force_extract: bool = False,
    return_parent_dir: bool = False,
    cache_dir=None,
    verbose: bool = False,
):
    """
    Attempts to cache a file or URL and return the path to the cached file.
    If required libraries 'cached_path' and 'gdown' are not installed, raises an ImportError.

    Args:
        url_or_filename (str): The URL or filename to be cached.
        extract_archive (bool, optional): Whether to extract the file if it's an archive. Defaults to False.
        force_extract (bool, optional): Whether to force extraction even if the destination already exists. Defaults to False.
        return_parent_dir (bool, optional): If True, returns the parent directory of the cached file. Defaults to False.
        cache_dir (str, optional): Directory to store cached files. Defaults to None.
        verbose (bool, optional): Whether to print informative messages during the process. Defaults to False.

    Raises:
        ImportError: If the required libraries 'cached_path' and 'gdown' are not imported.

    Returns:
        str: Path to the cached file or its parent directory, depending on the 'return_parent_dir' parameter.
    """
    if url_or_filename is None:
        return None
    if verbose:
        logger.info(
            f"caching path: {url_or_filename}, extract_archive: {extract_archive}, force_extract: {force_extract}, cache_dir: {cache_dir}"
        )

    try:
        if url_or_filename.startswith("gd://"):
            _path = cached_gdown(
                url_or_filename,
                verbose=verbose,
                extract_archive=extract_archive,
                force_extract=force_extract,
                cache_dir=cache_dir,
            )
        else:
            if _cpath is None or gdown is None:
                raise ImportError(
                    "Error importing required libraries 'cached-path'. "
                    "Please install them using 'pip install cached-path' and try again."
                )

            if cache_dir is None:
                cache_dir = Path.home() / ".hyfi" / ".cache" / "cached_path"
            else:
                cache_dir = Path(cache_dir) / "cached_path"

            _path = _cpath.cached_path(
                url_or_filename,
                extract_archive=extract_archive,
                force_extract=force_extract,
                cache_dir=cache_dir,
            ).as_posix()

        logger.info(f"cached path: {_path}")

        if Path(_path).is_file():
            _parent_dir = Path(_path).parent
        elif Path(_path).is_dir():
            _parent_dir = Path(_path)
        else:
            logger.warning(f"Unknown path: {_path}")
            return None

        return _parent_dir.as_posix() if return_parent_dir else _path
    except Exception as e:
        logger.error(e)
        return None


def cached_gdown(
    url, verbose=False, extract_archive=None, force_extract=False, cache_dir=None
):
    """
    :type url: str
          ex) gd://id:path
    :type verbose: bool
    :type extract_archive: bool
    :type force_extract: bool
    :type cache_dir: str
    :returns: str
    """
    if gdown is None:
        raise ImportError(
            "Error importing required libraries 'gdown'. "
            "Please install them using 'pip install gdown' and try again."
        )

    if verbose:
        logger.info(f"Downloading {url}...")
    if cache_dir is None:
        cache_dir = Path.home() / ".hyfi" / ".cache" / "gdown"
    else:
        cache_dir = Path(cache_dir) / "gdown"
    cache_dir.mkdir(parents=True, exist_ok=True)

    gd_prefix = "gd://"
    if url.startswith(gd_prefix):
        url = url[len(gd_prefix) :]
        _url = url.split(":")
        if len(_url) == 2:
            id_, path = _url
        else:
            id_ = _url[0]
            path = id_

        # If we're using the path!c/d/file.txt syntax, handle it here.
        fname = None
        extraction_path = path
        exclamation_index = path.find("!")
        if extract_archive and exclamation_index >= 0:
            extraction_path = path[:exclamation_index]
            fname = path[exclamation_index + 1 :]

        cache_path = cache_dir / f".{id_}" / extraction_path
        cache_path.parent.mkdir(parents=True, exist_ok=True)

        cache_path = gdown.cached_download(
            id=id_,
            path=cache_path.as_posix(),
            quiet=not verbose,
        )

        if extract_archive:
            extraction_path, files = extractall(cache_path, force_extract=force_extract)

            if not fname or not files:
                return extraction_path
            for f in files:
                if f.endswith(fname):
                    return f
        return cache_path

    else:
        logger.warning(f"Unknown url: {url}")
        return None


def extractall(path, to=None, force_extract=False):
    """Extract archive file.

    Parameters
    ----------
    path: str
        Path of archive file to be extracted.
    to: str, optional
        Directory to which the archive file will be extracted.
        If None, it will be set to the parent directory of the archive file.
    """
    import tarfile
    from zipfile import ZipFile

    if to is None:
        to = os.path.dirname(path)

    if path.endswith(".zip"):
        opener, mode = ZipFile, "r"
    elif path.endswith(".tar"):
        opener, mode = tarfile.open, "r"
    elif path.endswith(".tar.gz") or path.endswith(".tgz"):
        opener, mode = tarfile.open, "r:gz"
    elif path.endswith(".tar.bz2") or path.endswith(".tbz"):
        opener, mode = tarfile.open, "r:bz2"
    else:
        logger.warning(
            f"Could not extract '{path}' as no appropriate extractor is found"
        )
        return path, None

    def namelist(f):
        return f.namelist() if isinstance(f, ZipFile) else [m.path for m in f.members]

    def filelist(f):
        files = []
        for fname in namelist(f):
            fname = os.path.join(to, fname)
            files.append(fname)
        return files

    extraction_name = Path(path).stem
    extraction_path = f"{to}/{extraction_name}"
    if extraction_path and (
        os.path.isdir(extraction_path)
        and os.listdir(extraction_path)
        and not force_extract
    ):
        files = [
            os.path.join(dirpath, filename)
            for dirpath, _, filenames in os.walk(extraction_path)
            for filename in filenames
        ]

        return to, files

    with opener(path, mode) as f:
        f.extractall(path=to)

    return to, filelist(f)
